@@ -1,3 +1,14 @@
+/* A utility program originally written for the Linux OS SCSI subsystem
+*  Copyright (C) 2003-2012 D. Gilbert
+*  This program is free software; you can redistribute it and/or modify
+*  it under the terms of the GNU General Public License as published by
+*  the Free Software Foundation; either version 2, or (at your option)
+*  any later version.
+
+   This program issues the SCSI SEND DIAGNOSTIC command and in one case
+   the SCSI RECEIVE DIAGNOSTIC command to list supported diagnostic pages.
+*/
+
 #include <unistd.h>
 #include <fcntl.h>
 #include <stdio.h>
@@ -13,18 +24,7 @@
 #include "sg_cmds_basic.h"
 #include "sg_cmds_extra.h"
 
-/* A utility program for the Linux OS SCSI generic ("sg") device driver.
-*  Copyright (C) 2003-2007 D. Gilbert
-*  This program is free software; you can redistribute it and/or modify
-*  it under the terms of the GNU General Public License as published by
-*  the Free Software Foundation; either version 2, or (at your option)
-*  any later version.
-
-   This program issues the SCSI SEND DIAGNOSTIC command and in one case
-   the SCSI RECEIVE DIAGNOSTIC command to list supported diagnostic pages.
-*/
-
-static char * version_str = "0.34 20070714";
+static char * version_str = "0.38 20120919";
 
 #define ME "sg_senddiag: "
 
@@ -345,7 +345,7 @@ static int process_cl(struct opts_t * optsp, int argc, char * argv[])
 
 /* Return of 0 -> success, otherwise see sg_ll_send_diag() */
 static int do_senddiag(int sg_fd, int sf_code, int pf_bit, int sf_bit,
-                       int devofl_bit, int unitofl_bit, void * outgoing_pg, 
+                       int devofl_bit, int unitofl_bit, void * outgoing_pg,
                        int outgoing_len, int noisy, int verbose)
 {
     int long_duration = 0;
@@ -397,6 +397,7 @@ static int build_diag_page(const char * inp, unsigned char * mp_arr,
     unsigned int h;
     const char * lcp;
     char * cp;
+    char * c2p;
 
     if ((NULL == inp) || (NULL == mp_arr) ||
         (NULL == mp_arr_len))
@@ -470,7 +471,7 @@ static int build_diag_page(const char * inp, unsigned char * mp_arr,
         }
         *mp_arr_len = off;
     } else {        /* hex string on command line */
-        k = strspn(inp, "0123456789aAbBcCdDeEfF,");
+        k = strspn(inp, "0123456789aAbBcCdDeEfF, ");
         if (in_len != k) {
             fprintf(stderr, "build_diag_page: error at pos %d\n", k + 1);
             return 1;
@@ -484,8 +485,13 @@ static int build_diag_page(const char * inp, unsigned char * mp_arr,
                 }
                 mp_arr[k] = h;
                 cp = strchr(lcp, ',');
+                c2p = strchr(lcp, ' ');
+                if (NULL == cp)
+                    cp = c2p;
                 if (NULL == cp)
                     break;
+                if (c2p && (c2p < cp))
+                    cp = c2p;
                 lcp = cp + 1;
             } else {
                 fprintf(stderr, "build_diag_page: error at pos %d\n",
@@ -524,9 +530,10 @@ static struct page_code_desc pc_desc_arr[] = {
         {0xd, "Supported SES diagnostic pages (SES-2)"},
         {0xe, "Download microcode diagnostic pages (SES-2)"},
         {0xf, "Subenclosure nickname diagnostic pages (SES-2)"},
-        {0x3f, "Protocol specific SAS (SAS-1)"},
+        {0x3f, "Protocol specific (SAS transport)"},
         {0x40, "Translate address (direct access)"},
         {0x41, "Device status (direct access)"},
+        {0x42, "Rebuild assist (direct access)"},
 };
 
 static const char * find_page_code_desc(int page_num)
@@ -553,7 +560,7 @@ static void list_page_codes()
     printf("Page_Code  Description\n");
     for (k = 0; k < num; ++k, ++pcdp)
         printf(" 0x%02x      %s\n", pcdp->page_code,
-               (pcdp->desc ? pcdp->desc : "<unknown>"));   
+               (pcdp->desc ? pcdp->desc : "<unknown>"));
 }
 
 
@@ -585,6 +592,10 @@ int main(int argc, char * argv[])
     }
 
     if (NULL == opts.device_name) {
+        if (opts.do_list) {
+            list_page_codes();
+            return 0;
+        }
         fprintf(stderr, "No DEVICE argument given\n");
         if (opts.opt_new)
             usage();
@@ -605,7 +616,7 @@ int main(int argc, char * argv[])
             return SG_LIB_SYNTAX_ERROR;
         }
     }
-    
+
     if ((opts.do_doff || opts.do_uoff) && (! opts.do_deftest)) {
         if (opts.opt_new) {
             printf("setting --doff or --uoff only useful when -t is set\n");
@@ -649,18 +660,6 @@ int main(int argc, char * argv[])
                        "'-raw='\n");
         }
     }
-    if (NULL == opts.device_name) {
-        if (opts.do_list) {
-            list_page_codes();
-            return 0;
-        }
-        fprintf(stderr, "No DEVICE argument given\n");
-        if (opts.opt_new)
-            usage();
-        else
-            usage_old();
-        return SG_LIB_SYNTAX_ERROR;
-    }
 
     if ((sg_fd = sg_cmds_open_device(opts.device_name, 0 /* rw */,
                                      opts.do_verbose)) < 0) {
@@ -677,7 +676,7 @@ int main(int argc, char * argv[])
                 int secs;
 
                 secs = (rsp_buff[18] << 8) + rsp_buff[19];
-#ifdef SG3_UTILS_MINGW
+#ifdef SG_LIB_MINGW
                 printf("Expected extended self-test duration=%d seconds "
                        "(%g minutes)\n", secs, secs / 60.0);
 #else
